import logging

class SimpleFormatter(logging.Formatter):
    grey = "\x1b[248m"

    blue = "\x1b[36m"
    green = "\x1b[32m"
    brown = "\x1b[96m"
    red = "\x1b[90m"
    bold_red = "\x1b[196m"

    reset = "\x1b[0m"
    metadata_format = "[%(asctime)s] [%(levelname)s] [%(name)s] [%(filename)s:%(lineno)d] "
    message_format = "%(message)s"
    time_format = "%Y-%m-%dT%H:%M:%S"

    FORMATS = {
        logging.DEBUG: blue + metadata_format + reset + grey + message_format + reset,
        logging.INFO: green + metadata_format + reset + grey + message_format + reset,
        logging.WARNING: brown + metadata_format + reset + grey + message_format + reset,
        logging.ERROR: red + metadata_format + reset + grey + message_format + reset,
        logging.CRITICAL: bold_red + metadata_format + reset + grey + message_format + reset
    }

    def __init__(self, color = True, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.__color = color

    def format(self, record):
        if self.__color:
            log_fmt = self.FORMATS.get(record.levelno)
            formatter = logging.Formatter(fmt=log_fmt, datefmt=self.time_format)
        else:
            formatter = logging.Formatter(fmt=self.metadata_format + self.message_format, datefmt=self.time_format)

        return formatter.format(record)


class ConsoleHandler(logging.StreamHandler):
    def __init__(self):
        super().__init__()
        self.setFormatter(SimpleFormatter())
        self.setLevel(logging.DEBUG)



from .targets import *
from .file_helpers import *
from .build_image import build_image

__all__ = [
    # targets
    "PackagePriority", "UpstreamPackages",
    "CustomDeb",
    "HookScript",
    "StaticFile",
    "AptPreference", "AptPreferenceType",
    "DirectConfig",

    # file_helpers
    "render_template_to_file",
    "render_template_to_string",
    "escape_string_for_shell_script",
    "download_file",

    # build_image
    "build_image",
]


import logging
from pathlib import Path

from lbhelper.logger import ConsoleHandler

# ========== Auto Script Related ==========
AUTO_SCRIPT_DIR = Path("auto")
"""
Auto script directory under the image build directory.
"""

BUILD_SCRIPT_PATH = AUTO_SCRIPT_DIR / "build"
"""
``build`` script path under the auto script directory.
"""
CLEAN_SCRIPT_PATH = AUTO_SCRIPT_DIR / "clean"
"""
``clean`` script path under the auto script directory.
"""
CONFIG_SCRIPT_PATH = AUTO_SCRIPT_DIR / "config"
"""
``config`` script path under the auto script directory.
"""

# ========== Config Related ==========
CONFIG_DIR = Path('config')
"""
Configuration directory under the image build directory.
"""

# Static Files related
CHROOT_INCLUDE_DIR = CONFIG_DIR / "includes.chroot"
"""
Files presents in live/installed system
See https://live-team.pages.debian.net/live-manual/html/live-manual/customizing-contents.en.html#live-chroot-local-includes .
"""
BINARY_INCLUDE_DIR = CONFIG_DIR / "includes.binary"
"""
Files presents in iso and system. Similar to :obj:`CHROOT_INCLUDE_DIR`
"""

# Packages related
PACKAGE_LIST_DIR = CONFIG_DIR / "package-lists"
"""
Upstream package lists
See https://live-team.pages.debian.net/live-manual/html/live-manual/customizing-package-installation.en.html
"""
CHROOT_DEB_DIR = CONFIG_DIR / "packages.chroot"
"""
Custom deb directory
See https://live-team.pages.debian.net/live-manual/html/live-manual/customizing-package-installation.en.html#456
"""

# Package Preferences
BUILD_TIME_APT_PREFERENCE_FILE = CONFIG_DIR / "apt" / "preferences"
"""
Build time apt preference file
"""
RUN_TIME_APT_PREFERENCE_FILE = CHROOT_INCLUDE_DIR / "etc" / "apt" / "preferences"
"""
Run time apt preference file
"""

# Hooks
HOOKS_DIR = CONFIG_DIR / "hooks"
"""
Hooks base directory
See https://live-team.pages.debian.net/live-manual/html/live-manual/customizing-contents.en.html#539
"""
LIVE_HOOKS_DIR = HOOKS_DIR / "live"
"""
Live hooks directory
"""
NORMAL_HOOKS_DIR = HOOKS_DIR / "normal"
"""
Normal hooks directory
"""

# Bootloaders
BOOTLOADER_DIR = CONFIG_DIR / "bootloaders"
"""
Live build image bootloader directory
"""
BUILTIN_BOOTLOADER_DIR = Path("/usr/share/live/build/bootloaders")
"""
Default builtin bootloader directory on host. Copied as default bootloaders for live system.
See https://live-team.pages.debian.net/live-manual/html/live-manual/customizing-binary.en.html#640
"""

DEFAULT_LOGGER = logging.getLogger("lbhelper")
"""
Default logger
"""
DEFAULT_LOGGER.setLevel(logging.INFO)
DEFAULT_LOGGER.addHandler(ConsoleHandler())

DEFAULT_ISO_BUILD_DIR = Path("/tmp/iso")
"""
Default image build directory. All configs and files will be copied to here.
"""
DEFAULT_LIVE_BUILD_BINARY = Path("/usr/bin/lb")
"""
Default live build executable path
"""
DEFAULT_DPKG_NAME_BINARY = Path("/usr/bin/dpkg-name")
"""
Default dpkg-name executable path
"""

DEFAULT_IMAGE_NAME = "myos"
"""
Defaul built image name.
"""
DEFAULT_DISTRIBUTION = "trixie"
"""
Default debian distribution.
"""


from pathlib import Path

from .auto_scripts.auto_scripts import AutoScriptType
from .auto_scripts.auto_scripts import write_auto_script
from .defaults import DEFAULT_ISO_BUILD_DIR
from .defaults import DEFAULT_LOGGER
from .defaults import DEFAULT_DISTRIBUTION, DEFAULT_IMAGE_NAME
from .live_build import LBOperation
from .live_build import remove_build_dir
from .live_build import run_lb_operation
from .target_writer import AptPreferencesWriter
from .target_writer import CustomDebWriter
from .target_writer import DirectConfigWriter
from .target_writer import HookScriptWriter
from .target_writer import StaticFileWriter
from .target_writer import TargetWriter
from .target_writer import UpstreamPackagesWriter
from .targets import AptPreference
from .targets import CustomDeb
from .targets import DirectConfig
from .targets import HookScript
from .targets import StaticFile
from .targets import UpstreamPackages
from .extensions import copy_bootloaders


def build_image(
    targets: list,
    iso_build_dir: Path = DEFAULT_ISO_BUILD_DIR, fresh_build: bool = True,
    distribution: str = DEFAULT_DISTRIBUTION,
    image_name: str = DEFAULT_IMAGE_NAME,
    skip_build: bool = False,
):
    """
    Build image from targets.

    :param List targets: Targets to process. Elements can be :class:`Target` or List[:class:`Target`].
    :param pathlib.Path,Optional iso_build_dir: Image build directory.
    :param bool fresh_build: If ``True``, remove build directory before performing any operations for fresh build.
    :param str distribution: Base Debian distribution for image. ``trixie`` by default.
    :param str image_name: Name of image. ``myos`` by default.
    :param bool skip_build: If ``True``, ``lb build`` won't run.
    """
    DEFAULT_LOGGER.info(f"Cleanup build directory {iso_build_dir}")
    remove_build_dir(iso_build_dir=iso_build_dir, exist_ok=fresh_build)

    DEFAULT_LOGGER.info(f"Write auto-scripts")
    write_auto_script(
        iso_build_dir=iso_build_dir,
        script_type=AutoScriptType.CONFIG,
        distribution=distribution, image_name=image_name)
    write_auto_script(iso_build_dir=iso_build_dir, script_type=AutoScriptType.BUILD)
    write_auto_script(iso_build_dir=iso_build_dir, script_type=AutoScriptType.CLEAN)

    DEFAULT_LOGGER.info(f"Run lb config")
    run_lb_operation(operation=LBOperation.CONFIG, iso_build_dir=iso_build_dir)

    DEFAULT_LOGGER.info(f"Attach built-in targets")
    copy_bootloaders_target = copy_bootloaders(iso_build_dir=iso_build_dir)
    targets.append(copy_bootloaders_target)

    DEFAULT_LOGGER.info(f"Write targets")
    TargetWriter(
        target_handlers={
            UpstreamPackages: UpstreamPackagesWriter(iso_build_dir=iso_build_dir),
            CustomDeb: CustomDebWriter(iso_build_dir=iso_build_dir),
            HookScript: HookScriptWriter(iso_build_dir=iso_build_dir),
            StaticFile: StaticFileWriter(iso_build_dir=iso_build_dir),
            AptPreference: AptPreferencesWriter(iso_build_dir=iso_build_dir),
            DirectConfig: DirectConfigWriter(iso_build_dir=iso_build_dir),
        },
        targets=targets,
    ).execute()

    DEFAULT_LOGGER.info(f"Build image")
    if not skip_build:
        run_lb_operation(operation=LBOperation.BUILD, iso_build_dir=iso_build_dir)
    else:
        DEFAULT_LOGGER.info(f"Skip build")



import enum
from importlib.resources import files
from pathlib import Path

from lbhelper import defaults
from lbhelper.defaults import DEFAULT_LOGGER
from lbhelper.file_helpers import render_template_to_file
from dataclasses import dataclass


class AutoScriptType(enum.Enum):
    """
    Auto script type.
    """
    CLEAN = "clean"
    """
    Clean script.
    """
    BUILD = "build"
    """
    Build script.
    """
    CONFIG = "config"
    """
    Config script.
    """


@dataclass
class __AutoScriptInfo:
    """
    Information about an auto script. See :function:`write_auto_script`.

    :param pathlib.Path default_source_script_path: Builtin source script path. See the `auto`
       directory under the module.
    :param pathlib.Path target_script_path: Target script path.
    """
    default_source_script_path: Path
    target_script_path: Path


def write_auto_script(iso_build_dir: Path, script_type: AutoScriptType, source_script_path: Path = None, **kwargs):
    """
    Write specified auto script to :const:`defaults.AUTO_SCRIPT_PATH`. Target script path is predefined in
    this function. If ``source_script_path`` is not given, then builtin template will be used. Note that for the builtin
    config auto script, the ``distribution (str)`` and ``image_name (str)`` are required.

    :param pathlib.Path iso_build_dir: Image build directory.
    :param AutoScriptType script_type: Which type of auto script to write.
    :param pathlib.Path,Optional source_script_path: Source script template. If not given, then builtin
       auto script template mapped to the given ``script_type`` will be used.
    :param kwargs: Additional kwargs to render the template.
    """

    script_info = __get_script_info(iso_build_dir, script_type)

    source_script_path = script_info.default_source_script_path if source_script_path is None else source_script_path

    DEFAULT_LOGGER.info(f"Writing auto script '{script_type}' ...")
    render_template_to_file(source_script_path, script_info.target_script_path, **kwargs)
    DEFAULT_LOGGER.info(f"Auto script {script_type} saved.")


def __get_script_info(iso_build_dir: Path, script_type: AutoScriptType) -> __AutoScriptInfo:
    """
    Get script info for given auto script type.

    :param pathlib.Path iso_build_dir: Image build directory.
    :param AutoScriptType script_type: Auto script type.
    :return: auto script information
    :rtype: __AutoScriptInfo
    """
    module_path = files(__package__)

    script_info = {
        AutoScriptType.CONFIG: __AutoScriptInfo(
            default_source_script_path=Path(module_path / defaults.CONFIG_SCRIPT_PATH),
            target_script_path=iso_build_dir / defaults.CONFIG_SCRIPT_PATH,
        ),
        AutoScriptType.BUILD: __AutoScriptInfo(
            default_source_script_path=Path(module_path / defaults.BUILD_SCRIPT_PATH),
            target_script_path=iso_build_dir / defaults.BUILD_SCRIPT_PATH,
        ),
        AutoScriptType.CLEAN: __AutoScriptInfo(
            default_source_script_path=Path(module_path / defaults.CLEAN_SCRIPT_PATH),
            target_script_path=iso_build_dir / defaults.CLEAN_SCRIPT_PATH,
        )
    }

    return script_info[script_type]



from pathlib import Path
from lbhelper import defaults, DirectConfig


def copy_bootloaders(iso_build_dir: Path, source_bootloader_dir: Path = defaults.BUILTIN_BOOTLOADER_DIR) -> DirectConfig:
    """
    Copying bootloaders to image build directory. Each subdirectory under the source directory
    represents a bootloader.
    See https://live-team.pages.debian.net/live-manual/html/live-manual/customizing-binary.en.html#641

    :param iso_build_dir: Image build directory.
    :param source_bootloader_dir: Source bootloaders directory. It should contain 1 to multiple bootloaders.
    :return: bootloader copying target
    :rtype: DirectConfig
    """
    if not source_bootloader_dir.exists() or not source_bootloader_dir.is_dir():
        raise Exception(
            f"Bootloader directory {source_bootloader_dir} does not exist or is not a directory"
        )

    bootloader_dir = iso_build_dir.joinpath(defaults.BOOTLOADER_DIR)

    return DirectConfig(
        target_filepath=bootloader_dir,
        get_source_file=lambda :source_bootloader_dir
    )



import shutil
from abc import abstractmethod, ABCMeta
from pathlib import Path
from typing import Dict, List

import lbhelper.defaults as defaults
from . import AptPreferenceType
from .run_command import run_command
from .targets import Target, UpstreamPackages, PackagePriority, CustomDeb, HookScript, \
    StaticFile, AptPreference, DirectConfig


class TargetHandler(metaclass=ABCMeta):
    """
    Abstract class for target handlers.
    Must implement the execute method which called by the :class:`TargetWriter`.
    """
    @abstractmethod
    def execute(self, targets: List[Target]):
        """
        Called by the :class:`TargetWriter` as the entrypoint for running main target logic.
        See ``target_handlers`` parameter of :class:`TargetWriter`.

        :param List[Target] targets: List of target to process.
        """
        pass


class TargetWriter:
    """
    Process given targets based on corresponding handler.

    :param Dict[type(Target)] target_handlers: Map of target type to target handlers.
    :param List targets: Targets to process. Elements can be :class:`Target` or List[:class:`Target`].
    """
    def __init__(self, target_handlers: Dict[type(Target), TargetHandler.execute], targets: List):
        self.__targets = targets
        self.__target_handlers = target_handlers

        self.__collected_targets = {target_type: [] for target_type in target_handlers.keys()}

    def execute(self):
        """
        Process given targets based on corresponding handler.
        """
        defaults.DEFAULT_LOGGER.info("Write targets ... ")
        self.__collect_targets(self.__targets)
        self.__write_targets()

    def __collect_targets(self, targets: List):
        defaults.DEFAULT_LOGGER.info("Collecting targets ...")

        for target in targets:
            target_type = type(target)
            if target_type in self.__collected_targets.keys():
                self.__collected_targets[target_type].append(target)

            # if target is a list, then call self.__collect_target to collect targets in it
            elif target_type == list:
                self.__collect_targets(target)
            else:
                raise Exception(f"Unknown target type of target {target}: {target_type}")

        defaults.DEFAULT_LOGGER.info("All targets are collected.")

    def __write_targets(self):
        defaults.DEFAULT_LOGGER.info("Writing collected targets ...")
        for target_type, handler in self.__target_handlers.items():
            handler(self.__collected_targets[target_type])
        defaults.DEFAULT_LOGGER.info("All collected targets are saved.")


class UpstreamPackagesWriter(TargetHandler):
    """
    Write upstream packages to :const:`lbhelper.defaults.PACKAGE_LIST_DIR`.
    
    :param pathlib.Path,Optional iso_build_dir: Image build directory.
    """
    def __init__(self, iso_build_dir: Path):
        self.__package_list_dir = iso_build_dir.joinpath(defaults.PACKAGE_LIST_DIR)

    def execute(self, targets: List[UpstreamPackages]):
        defaults.DEFAULT_LOGGER.info("Writing upstream packages ...")

        for target in targets:
            self.__package_list_dir.mkdir(parents=True, exist_ok=True)
            filename = f"{target.package_set_code}.list.chroot" if not target.live_only else f"{target.package_set_code}.list.chroot_live"
            file_path = self.__package_list_dir / filename

            defaults.DEFAULT_LOGGER.info(f"Writing upstream packages {target.package_set_code} to {file_path}...")
            with open(file_path, "a") as f:
                if target.priority != PackagePriority.NOTSET:
                    f.write(f"! Packages Priority {target.priority}\n\n")

                for package in target.packages:
                    f.write(f"{package}\n")
            defaults.DEFAULT_LOGGER.info(f"Upstream packages {file_path} saved.")

        defaults.DEFAULT_LOGGER.info("All upstream packages saved.")


class CustomDebWriter(TargetHandler):
    """
    Save custom debs to :const:`lbhelper.defaults.CHROOT_DEB_DIR`.
    
    :param pathlib.Path,Optional iso_build_dir: Image build directory.
    :param pathlib.Path,Optional dpkg_name_binary: dpkg-name binary path. Used to correct ``.deb`` file name.
    """
    def __init__(self, iso_build_dir: Path, dpkg_name_binary: Path = defaults.DEFAULT_DPKG_NAME_BINARY):
        self.__chroot_deb_dir = iso_build_dir.joinpath(defaults.CHROOT_DEB_DIR)
        self.__dpkg_name_binary = dpkg_name_binary

    def execute(self, targets: List[CustomDeb]):
        defaults.DEFAULT_LOGGER.info("Writing custom deb ...")

        for target in targets:
            self.__chroot_deb_dir.mkdir(parents=True, exist_ok=True)
            source_deb = target.get_deb()
            temp_deb_path = self.__chroot_deb_dir / source_deb.name
            defaults.DEFAULT_LOGGER.info(f"Copying custom deb {source_deb} to {temp_deb_path}")
            shutil.copy(source_deb, temp_deb_path)
            run_command([str(self.__dpkg_name_binary), str(temp_deb_path)])

        defaults.DEFAULT_LOGGER.info("All custom deb saved.")


class AptPreferencesWriter(TargetHandler):
    """
    Add time apt preferences to :const:`lbhelper.defaults.BUILD_TIME_APT_PREFERENCE_FILE`
    / :const:`lbhelper.defaults.RUN_TIME_APT_PREFERENCE_FILE` depends on preference type.
    
    :param pathlib.Path,Optional iso_build_dir: Image build directory.
    """
    def __init__(self, iso_build_dir: Path):
        self.__build_time_apt_preference_file = iso_build_dir.joinpath(defaults.BUILD_TIME_APT_PREFERENCE_FILE)
        self.__run_time_apt_preference_file = iso_build_dir.joinpath(defaults.RUN_TIME_APT_PREFERENCE_FILE)

    def execute(self, targets: List[AptPreference]):
        defaults.DEFAULT_LOGGER.info("Writing build time apt preferences ...")
        self._write_apt_preferences(targets)
        defaults.DEFAULT_LOGGER.info("All build time apt preferences saved.")

    def _write_apt_preferences(self, targets: List[AptPreference]):
        for target in targets:
            self.__write_single_target(target)

    def __write_single_target(self, target: AptPreference):
        defaults.DEFAULT_LOGGER.info(f"Preference Type: {target.preference_type}")

        preference_file = self.__build_time_apt_preference_file \
            if target.preference_type is AptPreferenceType.BUILD_TIME \
            else self.__run_time_apt_preference_file

        preference_file.parent.mkdir(parents=True, exist_ok=True)

        defaults.DEFAULT_LOGGER.info(
            f"Writing apt preferences of {target.package} to {preference_file}.")
        with open(preference_file, "a") as f:
            f.writelines(
                [
                    f"Package: {target.package}\n",
                    f"Pin: {target.pin}\n",
                    f"Pin-Priority: {target.pin_priority}\n",
                    "\n",
                ]
            )
        defaults.DEFAULT_LOGGER.info(f"Apt preferences of {target.package} saved.")


class HookScriptWriter(TargetHandler):
    """
    Add hook scripts to :const:`lbhelper.defaults.LIVE_HOOKS_DIR` (for live system)
    and :const:`lbhelper.defaults.NORMAL_HOOKS_DIR` (for live/installed system).

    :param pathlib.Path,Optional iso_build_dir: Image build directory.
    """
    def __init__(self, iso_build_dir: Path):
        self.__live_hooks_dir = iso_build_dir.joinpath(defaults.LIVE_HOOKS_DIR)
        self.__normal_hooks_dir = iso_build_dir.joinpath(defaults.NORMAL_HOOKS_DIR)
        
        # hook priority of pre-defined hooks
        self.__builtin_live_hooks_orders = [10, 50]
        self.__builtin_normal_hooks_orders = [
            1000, 1010, 1020,
            5000, 5010, 5020, 5030, 5040, 5050,
            8000, 8010, 8020, 8030, 8040, 8050, 8060, 8070, 8080, 8090, 8100, 8110,
            9000, 9010, 9020
        ]

        self.__current_live_hook_order = 1
        self.__current_normal_hook_order = 1

    def execute(self, targets: List[HookScript]):
        defaults.DEFAULT_LOGGER.info(f"Writing hook scripts ...")

        for target in targets:
            if target.live_only:
                self.__write_single_target(target=target, hook_dir=self.__live_hooks_dir, current_order=self.__current_live_hook_order)
                self.__current_live_hook_order +=1
                if self.__current_live_hook_order in self.__builtin_live_hooks_orders:
                    self.__current_live_hook_order += 1
            else:
                self.__write_single_target(target=target, hook_dir=self.__normal_hooks_dir, current_order=self.__current_normal_hook_order)
                self.__current_normal_hook_order += 1
                if self.__current_normal_hook_order in self.__builtin_normal_hooks_orders:
                    self.__current_normal_hook_order += 1

        defaults.DEFAULT_LOGGER.info(f"All hook scripts saved.")

    @staticmethod
    def __write_single_target(target: HookScript, hook_dir: Path, current_order: int):
        hook_dir.mkdir(parents=True, exist_ok=True)
        hook_filename = f"{current_order:04}-{target.hook_name}.hook.chroot"
        source_path = target.get_script_file()
        target_path = hook_dir / hook_filename

        defaults.DEFAULT_LOGGER.info(f"Writing {source_path} to {target_path}")
        shutil.copy(source_path, target_path)
        defaults.DEFAULT_LOGGER.info(f"Hook {target_path} saved")


class StaticFileWriter(TargetHandler):
    """
    Add static file to :const:`lbhelper.defaults.CHROOT_INCLUDE_DIR` (for live/installed system)
    and :const:`lbhelper.defaults.BINARY_INCLUDE_DIR` (as part of iso).
    
    :param pathlib.Path,Optional iso_build_dir: Image build directory.
    """
    def __init__(self, iso_build_dir: Path):
        self.__chroot_include_dir = iso_build_dir.joinpath(defaults.CHROOT_INCLUDE_DIR)
        self.__binary_include_dir = iso_build_dir.joinpath(defaults.BINARY_INCLUDE_DIR)

    def execute(self, targets: List[StaticFile]):
        defaults.DEFAULT_LOGGER.info("Writing static files ...")

        for target in targets:
            include_dir = self.__binary_include_dir if target.binary_file else self.__chroot_include_dir
            target_file_path = include_dir.joinpath(target.target_filepath.relative_to(target.target_filepath.anchor))
            source_file = target.get_source_file()

            defaults.DEFAULT_LOGGER.info(f"Ensuring parent path of {target_file_path} ... ")
            target_file_path.parent.mkdir(parents=True, exist_ok=True)
            defaults.DEFAULT_LOGGER.info(f"Parent path {target_file_path.parent} created.")

            defaults.DEFAULT_LOGGER.info(f"Writing {source_file} to {target_file_path}")
            if source_file.is_file():
                shutil.copy(source_file, target_file_path)
            else:
                shutil.copytree(source_file, target_file_path, dirs_exist_ok=True, symlinks=True)
            defaults.DEFAULT_LOGGER.info(f"Static file {target_file_path} saved.")

        defaults.DEFAULT_LOGGER.info(f"All static files saved.")


class DirectConfigWriter(TargetHandler):
    """
    Add files to image build directory.

    :param pathlib.Path,Optional iso_build_dir: Image build directory.
    """
    def __init__(self, iso_build_dir: Path):
        self.__iso_build_dir = iso_build_dir

    def execute(self, targets: List[DirectConfig]):
        for target in targets:
            source_file = target.get_source_file()
            target_file_path = self.__iso_build_dir.joinpath(target.target_filepath.relative_to(target.target_filepath.anchor))

            defaults.DEFAULT_LOGGER.info(f"Ensuring parent path of {target_file_path} ... ")
            target_file_path.parent.mkdir(parents=True, exist_ok=True)

            defaults.DEFAULT_LOGGER.info(f"Writing {source_file} to {target_file_path}")
            if source_file.is_file():
                shutil.copy(source_file, target_file_path)
            else:
                shutil.copytree(source_file, target_file_path, dirs_exist_ok=True, symlinks=True)
            defaults.DEFAULT_LOGGER.info(f"Direct config {target_file_path} saved.")

        defaults.DEFAULT_LOGGER.info(f"All direct configs saved.")



import shutil
from enum import StrEnum
from pathlib import Path

from lbhelper import defaults
from lbhelper.run_command import run_command


def remove_build_dir(iso_build_dir: Path, exist_ok: bool = False):
    """
    Remove ``iso_build_dir``. Usually used before start fresh build.

    :param pathlib.Path iso_build_dir: Path to image build directory.
    :param bool exist_ok: If ``True``, remove ``iso_build_dir`` for fresh build.
    """
    if exist_ok:
        defaults.DEFAULT_LOGGER.info(f"Skip build dir {iso_build_dir} removal.")
        return

    defaults.DEFAULT_LOGGER.info(f"Removing build dir {iso_build_dir} ...")
    shutil.rmtree(iso_build_dir, ignore_errors=True)
    defaults.DEFAULT_LOGGER.info(f"Build dir {iso_build_dir} removed.")


class LBOperation(StrEnum):
    BUILD = "build"
    CLEAN = "clean"
    CONFIG = "config"


def run_lb_operation(
    operation: LBOperation,
    iso_build_dir: Path,
    live_build_binary: Path = defaults.DEFAULT_LIVE_BUILD_BINARY,
):
    """
    Run given live build operation in ``iso_build_dir``.

    :param LBOperation operation: A live build operation.
    :param pathlib.Path iso_build_dir: Image build directory.
    :param pathlib.Path,Optional live_build_binary: Live build binary path.
      Default to :const:`defaults.DEFAULT_LIVE_BUILD_BINARY`.
    """
    defaults.DEFAULT_LOGGER.info(f"Running lb {operation.value} in {iso_build_dir} ...")
    run_command(command=[str(live_build_binary), operation.value], cwd=iso_build_dir)
    defaults.DEFAULT_LOGGER.info(f"Finish {operation.value} in {iso_build_dir}.")



from dataclasses import dataclass
from enum import StrEnum
from pathlib import Path
from typing import List, Callable
from abc import ABCMeta


@dataclass
class Target(metaclass=ABCMeta):
    """
    Abstract class for all Targets.
    """
    pass


class PackagePriority(StrEnum):
    """
    Package priority. Used as value of :attr:`.UpstreamPackages.priority`.
    For more details about each value, see https://www.debian.org/doc/debian-policy/ch-archive.html#priorities .
    Note that `NOTSET` denotes not adding priority to result package set file.
    """
    REQUIRED = "required"
    IMPORTANT = "important"
    STANDARD = "standard"
    OPTIONAL = "optional"
    NOTSET = "NOTSET"


@dataclass
class UpstreamPackages(Target):
    """
    A set of upstream packages.

    :param List[str] packages: List of package names
    :param str package_set_code: A unique identifier for the package set.
    :param bool,Optional live_only: Only install in live image. Default is `False` (install in both live and installed system)
    :param PackagePriority,Optional priority: Package priority. `NOTSET` by default.
    """
    packages: List[str]
    package_set_code: str
    live_only: bool = False
    priority: PackagePriority = PackagePriority.NOTSET


@dataclass
class CustomDeb(Target):
    """
    A custom deb package.

    :param Callable[[], pathlib.Path] get_deb: Callback function to get source deb package.
    """

    get_deb: Callable[[], Path]


@dataclass
class HookScript(Target):
    """
    A hook runs in chroot time.
    See https://live-team.pages.debian.net/live-manual/html/live-manual/customizing-contents.en.html#539 .

    :param Callable[[], pathlib.Path] get_script_file: Callback function to get hook shell script.
    :param bool,Optional live_only: Only run in live image.
       Default is `False` (install in both live and installed system).
    """
    get_script_file: Callable[[], Path]
    hook_name: str
    live_only: bool = False


@dataclass
class StaticFile(Target):
    """
    A static file.

    :param pathlib.Path target_filepath: Expected path on built iso/system.
    :param Callable[[], pathlib.Path] get_source_file: Callback function to get source file.
    :param bool,Optional binary_file: Save file in chroot (`False`) to access via live/installed system.
        or as binary (`True`) to access in iso without booting system. `False` by default.
    """
    target_filepath: Path
    get_source_file: Callable[[], Path]
    binary_file: bool = False


@dataclass
class AptPreferenceType(StrEnum):
    """
    Apt preference type. Used as value of :attr:`AptPreference.preference_type`.
    See https://live-team.pages.debian.net/live-manual/html/live-manual/customizing-package-installation.en.html#514
    """
    BUILD_TIME = "build_time"
    RUN_TIME = "run_time"


@dataclass
class AptPreference(Target):
    """
    An apt preference.
    See https://live-team.pages.debian.net/live-manual/html/live-manual/customizing-package-installation.en.html#514
    and https://linux.die.net/man/5/apt_preferences .

    :param str package: Package selector.
    :param str pin: Pin preference.
    :param int pin_priority: Priority.
    :param AptPreferenceType preference_type: Apt preference type. Refer to :class:`AptPreferenceType`.
    """
    package: str
    pin: str
    pin_priority: int
    preference_type: AptPreferenceType


@dataclass
class DirectConfig(Target):
    """
    Directly copying files/directories to arbitrary destinations under an image build directory.
    Avoid directly using this target unless you have special requirements on customizing build
    process, which should be supported by extensions or new targets(PRs welcomed).

    :param pathlib.Path target_filepath: Expected path under the image build directory.
    :param Callable[[], pathlib.Path] get_source_file: Callback function to get source file.
    """
    target_filepath: Path
    get_source_file: Callable[[], Path]



from pathlib import Path
import tempfile
import jinja2
import requests

from lbhelper.defaults import DEFAULT_LOGGER


def render_template_to_file(template_path: Path, target_path: Path = None, **kwargs) -> Path:
    """
    Render template to a temp file and returns path of it. Generally used with ``pathlib.Path`` fields
    in :class:`Targets`.

    :param pathlib.Path template_path: Template path.
    :param pathlib.Path,Optional target_path: Target path. If not given, then writing to a temporary file.
    :param kwargs: Additional kwargs for template.
    :return: Path of rendered file.
    :rtype: pathlib.Path
    """

    DEFAULT_LOGGER.info(f"Rendering template {template_path} to {target_path} ...")

    rendered_template = render_template_to_string(template_path, **kwargs)
    written_file_path = target_path

    if target_path is None:
        DEFAULT_LOGGER.info(f"Rendering template to {target_path} ...")

        with tempfile.NamedTemporaryFile(mode="w+", delete=False) as tmp:
            tmp.write(rendered_template)
            written_file_path = Path(tmp.name)
    else:
        DEFAULT_LOGGER.info(f"Ensuring parent directory of {target_path} ...")
        target_path.parent.mkdir(parents=True, exist_ok=True)

        DEFAULT_LOGGER.info(f"Rendering template to {target_path} ...")
        with target_path.open(mode="w+", encoding="utf-8") as f:
            f.write(render_template_to_string(template_path, **kwargs))

    DEFAULT_LOGGER.info(f"Template {template_path} rendered to {target_path}.")

    return written_file_path

def render_template_to_string(template_path: Path, **kwargs) -> str:
    """
    Render template to string. Generally used as input as other templates. For example, when defining a
    :class:`HooKScript` to write a config file, the config file itself can be defined in another template
    and passed to the :class:`HookScript`, which decouples the :class:`HookScript` and the config file.

    :param pathlib.Path template_path: Template path.
    :param kwargs: Additional kwargs for template.
    :return: Rendered string.
    :rtype: str
    """
    env = jinja2.Environment(loader=jinja2.FileSystemLoader(template_path.parent), undefined=jinja2.StrictUndefined)
    template = env.get_template(template_path.name)
    return template.render(**kwargs)


def escape_string_for_shell_script(s: str) -> str:
    """
    Escape "\" and "$". Useful when writing string to file via shell scripts. For instance, when writing strings
    via:
    ```
    cat > file<<EOL
    $SOME_STRING
    EOL
    ```
    The "\" and "$" of  $SOME_STRING need to be escaped. Otherwise, they'll be treated as special characters from
    source script and cause unexpected results.

    :param str s: String to escape.
    :return: Escaped string.
    :rtype: str
    """
    patterns_dict = {
        "\\": "\\\\",
        "$": "\\$",
    }

    for pattern, alter in patterns_dict.items():
        s = s.replace(pattern, alter)

    return s


def download_file(url: str) -> Path:
    """
    Download file from url as temporary file and return path of it. Generally used with ``pathlib.Path`` fields
    in :class:`Targets`.

    :param str url: URL to download.
    :return: Path of downloaded file.
    :rtype: pathlib.Path
    """
    with requests.get(url, stream=True) as r:
        r.raise_for_status()
        with tempfile.NamedTemporaryFile(mode="wb", delete=False) as tmp:
            for chunk in r.iter_content(chunk_size=8192):
                tmp.write(chunk)
    return Path(tmp.name)



import subprocess
from pathlib import Path
from typing import List

from .defaults import DEFAULT_LOGGER


def run_command(command: List[str], cwd: Path = None):
    str_command = " ".join(c for c in command)

    DEFAULT_LOGGER.info(f"Executing command: \"{str_command}\"")
    process = subprocess.Popen(
        str_command,
        shell=True, cwd=cwd, stdout=subprocess.PIPE, stderr=subprocess.STDOUT
    )
    while process.poll() is None:
        write_command_to_logger(process)

    return_code = process.poll()

    DEFAULT_LOGGER.info(f"Command \"{str_command}\" exit with return code {return_code}")

    if return_code != 0:
        raise subprocess.CalledProcessError(returncode=return_code, cmd=str_command)


def write_command_to_logger(process: subprocess.Popen):
    while True:
        output = process.stdout.readline().decode()
        if output and len(output) > 0:
            if output[-1] == '\n':
               output = output[:-1]

            DEFAULT_LOGGER.info(output)
        else:
            break
